/*
   Copyright (C) 2008 Norbert Frese

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License (LGPL) as published by the Free Software Foundation;
   either version 2 of the License, or (at your option) any later
   version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301, USA.
*/

#include "giobridge.h"

//#include <config.h>

#include <QtCore/QBool> //for Q_OS_XXX
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/socket.h>
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#include <assert.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <grp.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <time.h>
#include <utime.h>
#include <unistd.h>
#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include <QtCore/QByteRef>
#include <QtCore/QDate>
#include <QtCore/QVarLengthArray>
#include <QtCore/QCoreApplication>
#include <QtCore/QRegExp>
#include <QtCore/QFile>

#include <kdebug.h>
#include <kurl.h>
#include <kcomponentdata.h>
#include <kconfig.h>
#include <kconfiggroup.h>
#include <ktemporaryfile.h>
#include <klocale.h>
#include <limits.h>
#include <kshell.h>
#include <kmountpoint.h>
#include <kstandarddirs.h>

#include <kio/ioslave_defaults.h>
#include <kde_file.h>
#include <kglobal.h>
#include <kmimetype.h>

#include <glib.h>
#include <gio/gio.h>

using namespace KIO;

#define MAX_IPC_SIZE (1024*32)

static QString testLogFile( const QByteArray&_filename );

extern "C" int KDE_EXPORT kdemain( int argc, char **argv )
{
  QCoreApplication app( argc, argv ); // needed for QSocketNotifier
  KComponentData componentData( "giobridge", "kdelibs4" );
  ( void ) KGlobal::locale();

  kDebug(7101) << "Starting giobridge " << getpid();

  if (argc != 4)
  {
     fprintf(stderr, "Usage: kio_file protocol domain-socket1 domain-socket2\n");
     exit(-1);
  }

  GioBridge slave(argv[2], argv[3]);
  slave.dispatchLoop();

  kDebug(7101) << "Done";
  return 0;
}


static int gIOError2KIO(GError * err, bool isFileOp)
{
  
  kDebug( 7101 ) << "gIOError2KIO! code=" << err->code << " msg=" << err->message << endl;
  
  int gioErrorCode= err->code;
  switch (gioErrorCode)
  {
    case G_IO_ERROR_FAILED              : return ERR_UNKNOWN;   
    case G_IO_ERROR_NOT_FOUND           : return ERR_DOES_NOT_EXIST;
    case G_IO_ERROR_EXISTS              : return isFileOp ? ERR_FILE_ALREADY_EXIST : ERR_DIR_ALREADY_EXIST;
    case G_IO_ERROR_IS_DIRECTORY        : return ERR_IS_DIRECTORY;
    case G_IO_ERROR_NOT_DIRECTORY       : return isFileOp ? ERR_UNKNOWN : ERR_IS_FILE;
    case G_IO_ERROR_NOT_EMPTY           : return ERR_COULD_NOT_RMDIR;
    case G_IO_ERROR_NOT_REGULAR_FILE    : return isFileOp ? ERR_IS_DIRECTORY : ERR_UNKNOWN;
    case G_IO_ERROR_NOT_SYMBOLIC_LINK   : return ERR_CANNOT_DELETE;  // todo
    case G_IO_ERROR_NOT_MOUNTABLE_FILE  : return ERR_COULD_NOT_MOUNT;
    case G_IO_ERROR_FILENAME_TOO_LONG   : return ERR_MALFORMED_URL;
    case G_IO_ERROR_INVALID_FILENAME    : return ERR_MALFORMED_URL;
    case G_IO_ERROR_TOO_MANY_LINKS      : return ERR_CYCLIC_LINK;
    case G_IO_ERROR_NO_SPACE            : return ERR_DISK_FULL;
    case G_IO_ERROR_INVALID_ARGUMENT    : return ERR_UNKNOWN;
    case G_IO_ERROR_PERMISSION_DENIED   : return ERR_ACCESS_DENIED;
    case G_IO_ERROR_NOT_SUPPORTED       : return ERR_UNSUPPORTED_ACTION;
    case G_IO_ERROR_NOT_MOUNTED         : return ERR_ACCESS_DENIED; // todo: new KDE error code
    case G_IO_ERROR_ALREADY_MOUNTED     : return ERR_UNSUPPORTED_ACTION; // todo: new KDE error code
    case G_IO_ERROR_CLOSED              : return ERR_CONNECTION_BROKEN; // todo
    case G_IO_ERROR_CANCELLED           : return ERR_USER_CANCELED;
    case G_IO_ERROR_PENDING             : return ERR_INTERNAL; // todo
    case G_IO_ERROR_READ_ONLY           : return ERR_UNKNOWN; // needs a better translation
    case G_IO_ERROR_CANT_CREATE_BACKUP  : return ERR_INTERNAL; // todo
    case G_IO_ERROR_WRONG_ETAG          : return ERR_UNKNOWN; // todo
    case G_IO_ERROR_TIMED_OUT           : return ERR_SERVER_TIMEOUT; 
    case G_IO_ERROR_WOULD_RECURSE       : return ERR_UNKNOWN; // todo
    case G_IO_ERROR_BUSY                : return ERR_INTERNAL;
    case G_IO_ERROR_WOULD_BLOCK         : return ERR_INTERNAL;
    case G_IO_ERROR_HOST_NOT_FOUND      : return ERR_SERVICE_NOT_AVAILABLE;
    case G_IO_ERROR_WOULD_MERGE         : return ERR_INTERNAL; // todo
    case G_IO_ERROR_FAILED_HANDLED      : return ERR_INTERNAL; // todo
  }
  return ERR_UNKNOWN;
}

static void gFileInfoListAttributes(GFileInfo * gFileInfo)
{
  char** attrs = g_file_info_list_attributes(gFileInfo, NULL);
  int cnt=0;
  while (attrs[cnt])
  {
    char * value =  g_file_info_get_attribute_as_string (gFileInfo,
        attrs[cnt]);
    kDebug(7101) << "attrname=" << attrs[cnt] << " value=" << value << endl;
    g_free(value);
    cnt++;
  }
  g_strfreev(attrs); 
}


static void gFileInfo2UDSEntry(GFileInfo * gFileInfo, 
                            const char * basename, UDSEntry & entry,
                            short int details, bool withACL )
{
  
    gFileInfoListAttributes(gFileInfo);
  
    assert(entry.count() == 0); // by contract :-)
  
    // file name
    
    entry.insert( KIO::UDSEntry::UDS_NAME, QString(basename));
    
    // access permissions
    
    mode_t access=0;
    
    //entry.insert( KIO::UDSEntry::UDS_ACCESS, access );
    
    gboolean access_can_read = g_file_info_get_attribute_boolean(gFileInfo,
        G_FILE_ATTRIBUTE_ACCESS_CAN_READ);

    gboolean access_can_write = g_file_info_get_attribute_boolean(gFileInfo,
        G_FILE_ATTRIBUTE_ACCESS_CAN_WRITE);

    gboolean access_can_execute = g_file_info_get_attribute_boolean(gFileInfo,
        G_FILE_ATTRIBUTE_ACCESS_CAN_EXECUTE);

    gboolean access_can_delete = g_file_info_get_attribute_boolean(gFileInfo,
        G_FILE_ATTRIBUTE_ACCESS_CAN_DELETE);
    
    gboolean access_can_trash = g_file_info_get_attribute_boolean(gFileInfo,
        G_FILE_ATTRIBUTE_ACCESS_CAN_TRASH);
    
    gboolean access_can_rename = g_file_info_get_attribute_boolean(gFileInfo,
        G_FILE_ATTRIBUTE_ACCESS_CAN_RENAME);
    
    
    if (access_can_read)
      access |= S_IREAD;
    
    if (access_can_write)
      access |= S_IWRITE;
    
    if (access_can_execute)
      access |= S_IEXEC;
    
    entry.insert( KIO::UDSEntry::UDS_ACCESS, access );
    
    // file size
    
    goffset gsize = g_file_info_get_size(gFileInfo);
    
    entry.insert( KIO::UDSEntry::UDS_SIZE, gsize);
    
    // modification time 

    GTimeVal gMTime;
    
    g_file_info_get_modification_time   (gFileInfo, &gMTime);
    
    entry.insert( KIO::UDSEntry::UDS_MODIFICATION_TIME,(long long int) gMTime.tv_sec * 1000 + gMTime.tv_usec / 1000);
    
    // access time
    
    if (g_file_info_has_attribute(gFileInfo, G_FILE_ATTRIBUTE_TIME_ACCESS))
    {
      guint64 gATime = g_file_info_get_attribute_uint64(gFileInfo, G_FILE_ATTRIBUTE_TIME_ACCESS);
      entry.insert( KIO::UDSEntry::UDS_ACCESS_TIME, gATime);
    }
    
    // user name

    const char * ownerUser = g_file_info_get_attribute_string(gFileInfo, G_FILE_ATTRIBUTE_OWNER_USER);    
    if (ownerUser != NULL)
    {
      entry.insert( KIO::UDSEntry::UDS_USER, ownerUser);
    }

    // group name
    
    const char * ownerGroup = g_file_info_get_attribute_string(gFileInfo, G_FILE_ATTRIBUTE_OWNER_GROUP);    
    if (ownerGroup != NULL)
    {
      entry.insert( KIO::UDSEntry::UDS_GROUP, ownerGroup);
    }
    
    // symlink target
    
    const char * symlinkTarget = g_file_info_get_symlink_target(gFileInfo);
    if (symlinkTarget != NULL)
    {
      entry.insert( KIO::UDSEntry::UDS_LINK_DEST, QString(symlinkTarget));
    }
    
    // target uri

    const char * targetUri = g_file_info_get_attribute_string(gFileInfo, G_FILE_ATTRIBUTE_STANDARD_TARGET_URI);    
    if (targetUri != NULL)
    {
      entry.insert( KIO::UDSEntry::UDS_LINK_DEST, QString(targetUri).append("/"));
      
    }
    

    // file type

     GFileType gFileType = g_file_info_get_file_type(gFileInfo);
     
     mode_t type;

     switch (gFileType)
     {
       case G_FILE_TYPE_UNKNOWN : type = 0; break;
       case G_FILE_TYPE_REGULAR : type = S_IFREG; break;
       case G_FILE_TYPE_DIRECTORY : type = S_IFDIR; break;
       case G_FILE_TYPE_SYMBOLIC_LINK : type = S_IFLNK; break; // S_IFREG ?
       case G_FILE_TYPE_SPECIAL : type = S_IFSOCK; break;
       case G_FILE_TYPE_SHORTCUT : type = S_IFDIR; break;
       case G_FILE_TYPE_MOUNTABLE : type = S_IFDIR; break;
       default : type = 0; break;
     }

     if (type != 0)
     {
       entry.insert( KIO::UDSEntry::UDS_FILE_TYPE, type );
     }    
    
    
    kDebug(7101) << "gFileInfo: basename=" << basename << " filetype=" << gFileType 
      << " canread=" << access_can_read << " canwrite=" << access_can_write 
      << " canexecute=" << access_can_execute << " candelete=" << access_can_delete
      << " cantrash=" << access_can_trash << " canrename=" << access_can_rename
      << " size=" << gsize << " gmtime.sec=" << gMTime.tv_sec 
      << " user=" << ownerUser << " group=" << ownerGroup
      << " symlinktarget=" << symlinkTarget 
      << " targetURL=" << targetUri
      << endl;
      
    return;
}



GioBridge::GioBridge( const QByteArray &pool, const QByteArray &app )
    : SlaveBase( "*", pool, app )
{
	  kDebug(7101) << "initializing GioBridge";
	  g_type_init ();
}

GioBridge::~GioBridge()
{
}


void GioBridge::chmod( const KUrl& url, int permissions )
{
    GError * err = NULL;
  
    kDebug(7101) << "GioBridge::chmod(): " << url << ", permissions = " << permissions;
    
    GFile * gFile = g_file_new_for_uri(url.url().toUtf8().data());
    
    GFileInfo* gFileInfo = g_file_info_new();  
    
    g_file_info_set_attribute_boolean (gFileInfo, G_FILE_ATTRIBUTE_ACCESS_CAN_READ, permissions & S_IREAD);    
    g_file_info_set_attribute_boolean (gFileInfo, G_FILE_ATTRIBUTE_ACCESS_CAN_WRITE, permissions & S_IWRITE);
    
    g_file_set_attributes_from_info(gFile, gFileInfo,
        G_FILE_QUERY_INFO_NONE, NULL, &err);
  
    if (err)
    {
      error( gIOError2KIO(err, true), err->message );
      g_error_free(err);
    }
    else
    {
      finished();
    }
    g_object_unref (gFileInfo);
    g_object_unref (gFile);    

}

void GioBridge::chown( const KUrl& url, const QString& owner, const QString& group )
{
    kDebug(7101) << "GioBridge::chown(): " << url << ", owner = " << owner << ", group = " << group;

    GError * err = NULL;    
    
    GFile * gFile = g_file_new_for_uri(url.url().toUtf8().data());
    
    GFileInfo* gFileInfo = g_file_info_new();  
    
    g_file_info_set_attribute_string(gFileInfo, G_FILE_ATTRIBUTE_OWNER_USER, owner.toAscii());    
    g_file_info_set_attribute_string(gFileInfo, G_FILE_ATTRIBUTE_OWNER_GROUP, group.toAscii());
    
    g_file_set_attributes_from_info(gFile, gFileInfo,
        G_FILE_QUERY_INFO_NONE, NULL, &err);

    if (err)
    {
      error( gIOError2KIO(err, true), err->message );
      g_error_free(err);
    }
    else
    {
      finished();
    }
    g_object_unref (gFileInfo);
    g_object_unref (gFile);    
}

void GioBridge::setModificationTime( const KUrl& url, const QDateTime& mtime )
{
    kDebug(7101) << "GioBridge::setModificationTime(): " << url << ", mtime = " << mtime;
  
    GError * err = NULL;
  
    GFile * gFile = g_file_new_for_uri(url.url().toUtf8().data());  
  
    guint64 gmtime = mtime.toTime_t ();
    
    g_file_set_attribute(gFile, G_FILE_ATTRIBUTE_TIME_MODIFIED,
        G_FILE_ATTRIBUTE_TYPE_UINT64,
        &gmtime,
        G_FILE_QUERY_INFO_NONE,
        NULL,
        &err);  
  
    if (err)
    {
      error( gIOError2KIO(err, true), err->message );
      g_error_free(err);
    }
    else
    {
      finished();
    }
    g_object_unref (gFile);
}

void GioBridge::mkdir( const KUrl& url, int permissions )
{
    kDebug(7101) << "GioBridge::mkdir(): " << url << ", permissions = " << permissions;

    GError * err = NULL;

    GFile * gFile = g_file_new_for_uri(url.url().toUtf8().data());  
  
    g_file_make_directory(gFile, NULL, &err);
  
    if (err)
    {
      error( gIOError2KIO(err, true), err->message );
      g_error_free(err);
    }
    else
    {
      finished();
    }
    g_object_unref (gFile);
  
    return;
}

void GioBridge::get( const KUrl& url )
{
    goffset gTotalSize=-1;
  
    kDebug(7101) << "GioBridge::get " << url.url();

    GError * err = NULL;
  
    GFile * gFile = g_file_new_for_uri(url.url().toUtf8().data());
    
    
    GFileInputStream* gFileInputStream = g_file_read(gFile,
                                                          NULL,
                                                          &err);    
    
    if (err)
    {
      error( gIOError2KIO(err, true), err->message );
      g_error_free(err);
      
      g_object_unref (gFile);
      if (gFileInputStream) g_object_unref (gFileInputStream);      
      return;
    }    
    
    // query fileinfo
    
    GFileInfo* gFileInfo = g_file_input_stream_query_info(gFileInputStream,
                                                             "*",
                                                             NULL,
                                                             &err);    
    if (err)
    {
      if (err->code != 15)
      {
        error( gIOError2KIO(err, true), err->message );
        g_error_free(err);
      
        g_object_unref (gFile);
        if (gFileInputStream) g_object_unref (gFileInputStream);
        return;
      }
      // query info unsupported -> stat 
      
      kDebug( 7101 ) << "g_file_query_info! " << endl;
      
      g_error_free(err);
      err = NULL;
      
      gFileInfo = g_file_query_info( gFile,
               "*",
               G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS,
               NULL,
               &err);
       
      if (err)
      {
         error( gIOError2KIO(err, true), err->message );
         g_error_free(err);
         
         g_object_unref (gFile);
         g_object_unref (gFileInputStream);
         if (gFileInfo) g_object_unref (gFileInfo);   
         return;
         
      }      
    }    
      
    const char* contentType = g_file_info_get_content_type(gFileInfo);
    if (contentType)
    {
      kDebug( 7101 ) << "content type: " << contentType << endl;
      emit mimeType(contentType);      
    }
    
    gTotalSize = g_file_info_get_size(gFileInfo);
    kDebug( 7101 ) << "total size: " << gTotalSize << endl;
    
    totalSize( gTotalSize );  // emit totalSize

    KIO::filesize_t processed_size = 0;
    
    // resume?
    
    QString resumeOffset = metaData("resume");
    if ( !resumeOffset.isEmpty() )
    {
        bool ok;
        KIO::fileoffset_t offset = resumeOffset.toLongLong(&ok);
        if (ok && (offset > 0) && (offset < gTotalSize))
        {
          gssize skipped = g_input_stream_skip(G_INPUT_STREAM(gFileInputStream),
                                               offset,
                                               NULL,
                                               &err); // use seek instead?          
          if (skipped==-1)
          {
            error( gIOError2KIO(err, true), err->message );
            g_error_free(err);
            
            g_object_unref (gFile);
            g_object_unref (gFileInfo);
            g_object_unref (gFileInputStream);      
            return;
          }
          
          canResume ();
          processed_size = offset;
          kDebug( 7101 ) << "Resume offset: " << KIO::number(offset);
        }
    }    
    
    

    char buffer[ MAX_IPC_SIZE ];
    QByteArray array;

    while( 1 )
    {
      
      gssize n = g_input_stream_read(G_INPUT_STREAM(gFileInputStream),
            buffer,
            MAX_IPC_SIZE,
            NULL,
            &err);
      
      if (n == -1)
      {
        error( gIOError2KIO(err, true), err->message );  // KIO::ERR_COULD_NOT_READ?
        g_object_unref (gFile);
        g_object_unref (gFileInputStream);         
        return;
      }
      if (n == 0)
        break; // Finished

      array = QByteArray::fromRawData(buffer, n);
      data( array );
      array.clear();

      processed_size += n;
      processedSize( processed_size );

      kDebug( 7101 ) << "Processed: " << KIO::number (processed_size);
    }

    data( QByteArray() );

    processedSize( gTotalSize );
    finished();
    
    g_object_unref (gFile);
    g_object_unref (gFileInfo);
    g_object_unref (gFileInputStream);      
    
}

void GioBridge::put( const KUrl& url, int _mode, KIO::JobFlags _flags )
{
    kDebug(7101) << "GioBridge::put: " << url << ", mode=" << _mode;
  
    GError * err = NULL;
  
    GFile * gFile = g_file_new_for_uri(url.url().toUtf8().data());
    
    GFileOutputStream* gFileOutputStream=NULL;
    
    if (!(_flags & KIO::Overwrite))
    {
        gFileOutputStream = g_file_create(gFile,
                    G_FILE_CREATE_NONE,
                    NULL,
                    &err);  
    }
    else
    {
        gFileOutputStream = g_file_replace(gFile,
                    NULL,
                    FALSE,
                    G_FILE_CREATE_NONE,
                    NULL,
                    &err);
    }

    if (err)
    {
      error( gIOError2KIO(err, true), err->message );
      g_error_free(err);
      g_object_unref (gFile);
      return;
    }    

    int result=0;
    // Loop writing
    do
    {
        QByteArray buffer;
        dataReq(); // Request for data
        result = readData( buffer );

        if (result >= 0)
        {
          gsize bytes_written=-1;
          
          g_output_stream_write_all(G_OUTPUT_STREAM(gFileOutputStream),
              buffer.data(),
              buffer.size(),
              &bytes_written,
              NULL,
              &err);          

          if (err)
          {
            error( gIOError2KIO(err, true), err->message );
            g_error_free(err);
            g_object_unref (gFile);
            g_object_unref (gFileOutputStream);
            return;
          }
          
          g_assert(bytes_written == buffer.size());
          
        }
    }
    while ( result > 0 );

    g_output_stream_close(G_OUTPUT_STREAM(gFileOutputStream), NULL, &err);    
    if (err)
    {
      error( gIOError2KIO(err, true), err->message );
      g_error_free(err);
      g_object_unref (gFile);
      g_object_unref (gFileOutputStream);
      return;
    }    
    
    g_object_unref (gFileOutputStream);

    // set permissions
    if ( _mode != -1)
    {
      GFileInfo* gFileInfo = g_file_info_new();  
      
      g_file_info_set_attribute_boolean (gFileInfo, G_FILE_ATTRIBUTE_ACCESS_CAN_READ, _mode & S_IREAD);    
      g_file_info_set_attribute_boolean (gFileInfo, G_FILE_ATTRIBUTE_ACCESS_CAN_WRITE, _mode & S_IWRITE);
      
      g_file_set_attributes_from_info(gFile, gFileInfo,
          G_FILE_QUERY_INFO_NONE, NULL, &err);
    
      if (err)
      {
        kDebug(7101) << "GioBridge::put set permissions "<<  err->message;
        g_error_free(err);
        err=NULL;
      }
      g_object_unref (gFileInfo);
    }

    // set modification time
    const QString mtimeStr = metaData( "modified" );
    if ( !mtimeStr.isEmpty() ) {
        QDateTime dt = QDateTime::fromString( mtimeStr, Qt::ISODate );
        if ( dt.isValid() ) {
          
          guint64 gmtime = dt.toTime_t(); // modification time
          
          g_file_set_attribute(gFile, G_FILE_ATTRIBUTE_TIME_MODIFIED,
              G_FILE_ATTRIBUTE_TYPE_UINT64,
              &gmtime,
              G_FILE_QUERY_INFO_NONE,
              NULL,
              &err);  
        
          if (err)
          {
            kDebug(7101) << "GioBridge::put set time modified "<<  err->message;
            g_error_free(err);
            err=NULL;
          }          
          
        }
    }

    g_object_unref (gFile);
    finished();
}

static void _my_GFileProgressCallback(goffset current_num_bytes,
                               goffset total_num_bytes,
                               gpointer user_data)
{
    GioBridge * bridge = (GioBridge * )user_data;
    
    bridge->processedSize( current_num_bytes );
    bridge->totalSize( total_num_bytes );
    
}


void GioBridge::copy( const KUrl &src, const KUrl &dest,
                         int _mode, JobFlags _flags )
{
  
    kDebug(7101) << "GioBridge::copy(): " << src << " --> " << dest;
  
    GError * err = NULL;
  
    GFile * gFileSource = g_file_new_for_uri(src.url().toUtf8().data());  
    GFile * gFileDest = g_file_new_for_uri(dest.url().toUtf8().data());
  
    int copyFlags = G_FILE_COPY_NONE;   
    if (_flags & KIO::Overwrite) copyFlags |= G_FILE_COPY_OVERWRITE; 
    
    g_file_copy(gFileSource, gFileDest,
        (GFileCopyFlags) copyFlags,
        NULL,
        _my_GFileProgressCallback,
        this,
        &err);
    
    if (err)
    {
      error( gIOError2KIO(err, true), err->message );
      g_error_free(err);
    }
    else
    {
      finished();
    }
    g_object_unref (gFileSource);
    g_object_unref (gFileDest);  

}

void GioBridge::rename( const KUrl &src, const KUrl &dest,
                           KIO::JobFlags _flags )
{
    kDebug(7101) << "GioBridge::rename(): " << src << " --> " << dest;
  
    GError * err = NULL;
  
    GFile * gFileSource = g_file_new_for_uri(src.url().toUtf8().data());  
    GFile * gFileDest = g_file_new_for_uri(dest.url().toUtf8().data());
  
    g_file_move(gFileSource, gFileDest,
        G_FILE_COPY_NONE,
        NULL,
        _my_GFileProgressCallback,
        this,
        &err);
    
    if (err)
    {
      error( gIOError2KIO(err, true), err->message );
      g_error_free(err);
    }
    else
    {
      finished();
    }
    g_object_unref (gFileSource);
    g_object_unref (gFileDest);
  
}

void GioBridge::symlink( const QString &target, const KUrl &dest, KIO::JobFlags flags )
{

    kDebug(7101) << "GioBridge::symlink(): " << target << " <-- " << dest;
  
    GError * err = NULL;
  
    GFile * gFileDest = g_file_new_for_uri(dest.url().toUtf8().data());
    
    g_file_make_symbolic_link(gFileDest,
              target.toUtf8 (),
              NULL,
              &err);
  
    if (err)
    {
      error( gIOError2KIO(err, true), err->message );
      g_error_free(err);
    }
    else
    {
      finished();
    }
    g_object_unref (gFileDest);
  
}

void GioBridge::del( const KUrl& url, bool isfile)
{
    kDebug( 7101 ) <<  "Deleting "<< url.url() << " isfile=" << isfile;
    
    GError * err = NULL;    
    
    GFile * gFile = g_file_new_for_uri(url.url().toUtf8().data());
    
    GFileInfo * gFileInfo = g_file_query_info( gFile,
            "*",
            G_FILE_QUERY_INFO_NONE,
            NULL,
            &err);
    
    if (err)
    {
      error( gIOError2KIO(err, true), err->message );
      g_error_free(err);
      
      g_object_unref (gFile);
      if (gFileInfo) g_object_unref (gFileInfo);   
      return;
      
    }
    
    GFileType gFileType = g_file_info_get_file_type(gFileInfo);
    if (gFileInfo) g_object_unref (gFileInfo);   
    
    if (isfile && gFileType != G_FILE_TYPE_REGULAR)
    {
      error( KIO::ERR_IS_DIRECTORY, url.url() );
      g_object_unref (gFile);
      return;
    }
    
    if (!isfile && gFileType != G_FILE_TYPE_DIRECTORY)
    {
      error( KIO::ERR_IS_FILE, url.url() );
      g_object_unref (gFile);
      return;
    }  
    
    g_file_delete(gFile, NULL, &err);
  
    if (err)
    {
      error( gIOError2KIO(err, true), err->message );
      g_error_free(err);
    }
    else
    {
      finished();
    }
    g_object_unref (gFile);
    
}



void GioBridge::stat( const KUrl & url )
{

    kDebug(7101) << "GioBridge::stat " << url.url();
    
    // detail level?    
    QString sDetails = metaData(QLatin1String("details"));
    int details = sDetails.isEmpty() ? 2 : sDetails.toInt();
    kDebug(7101) << "GioBridge::stat details=" << details;

    // stat file
    GError * err = NULL;
  
    GFile * gFile = g_file_new_for_uri(url.url().toUtf8().data());
    
    GFileInfo * gFileInfo = g_file_query_info( gFile,
            "*",
            details > 1 ? G_FILE_QUERY_INFO_NONE : G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS,
            NULL,
            &err);
    
    if (err)
    {
      error( gIOError2KIO(err, true), err->message );
      g_error_free(err);
      
      g_object_unref (gFile);
      if (gFileInfo) g_object_unref (gFileInfo);   
      return;
      
    }
    
    
    char * basename = g_file_get_basename(gFile);    
    
    // create UDS Entry
    UDSEntry entry;
    
    gFileInfo2UDSEntry(gFileInfo, basename, entry, details, true /*with acls*/ );    
    
    // push result
    
    statEntry( entry );

    finished();
    
    // free objects
    g_free(basename);
    g_object_unref (gFile);
    g_object_unref (gFileInfo);
}


void GioBridge::listDir( const KUrl& url)
{
    kDebug(7101) << "========= LIST " << url.url() << " =========";

    GError * err = NULL;    
    
    GFile * gFile = g_file_new_for_uri(url.url().toUtf8().data());

    GFileEnumerator * gFileEnumerator = g_file_enumerate_children(
         gFile,
         "*",
         G_FILE_QUERY_INFO_NONE,
         NULL,
         &err);
    
    if (err)
    {
      error( gIOError2KIO(err, false), err->message );
      g_error_free(err);      
      g_object_unref (gFile);
      g_object_unref (gFileEnumerator);
      return;
    }    

    UDSEntry entry;
    
    for(;;)
    {
    
      GFileInfo* gFileInfo = g_file_enumerator_next_file(gFileEnumerator,
                                      NULL,
                                      &err);

      if (err)
      {
        error( gIOError2KIO(err, false), err->message );
        g_error_free(err);      
        g_object_unref (gFile);
        g_object_unref (gFileEnumerator);
        g_object_unref (gFileInfo);
        return;
      }    
      
      if (!gFileInfo)
      {
        break;  // done! 
      }
      
      const char*  basename = g_file_info_get_name(gFileInfo);

      entry.clear();
      gFileInfo2UDSEntry(gFileInfo, basename, entry, 2, true);
      listEntry( entry, false);
      g_object_unref(gFileInfo);
    }

    listEntry( entry, true ); // ready

    kDebug(7101) << "============= COMPLETED LIST ============";

    finished();
    
    g_object_unref (gFile);
    g_object_unref (gFileEnumerator);
}

// Connected to KShred
void GioBridge::slotProcessedSize( KIO::filesize_t bytes )
{
    kDebug(7101) << "GioBridge::slotProcessedSize (" << (unsigned int) bytes << ")";
    processedSize( bytes );
}

// Connected to KShred
void GioBridge::slotInfoMessage( const QString & msg )
{
    kDebug(7101) << "GioBridge::slotInfoMessage (" << msg << ")";
    infoMessage( msg );
}


#include "giobridge.moc"
